#!/usr/bin/env python
# coding: utf-8

import time
import pybullet
import threading
import pybullet_data
from qibullet.camera import Camera
from qibullet.pepper_virtual import PepperVirtual


class SimulationManager:
    """
    Class allowing to handle the different parameters of a pybullet simulation
    """

    def __init__(self):
        """
        Constructor
        """
        pass

    def launchSimulation(self, gui=True):
        """
        Launches a simulation instance

        Parameters:
            gui - Boolean, if True the simulation is launched with a GUI, and
            with no GUI otherwise

        Returns:
            physics_client - The id of the simulation client created
        """
        if gui:
            physics_client = pybullet.connect(pybullet.GUI)
            pybullet.setRealTimeSimulation(1, physicsClientId=physics_client)
            pybullet.configureDebugVisualizer(
                pybullet.COV_ENABLE_RGB_BUFFER_PREVIEW,
                0,
                physicsClientId=physics_client)
            pybullet.configureDebugVisualizer(
                pybullet.COV_ENABLE_DEPTH_BUFFER_PREVIEW,
                0,
                physicsClientId=physics_client)
            pybullet.configureDebugVisualizer(
                pybullet.COV_ENABLE_SEGMENTATION_MARK_PREVIEW,
                0,
                physicsClientId=physics_client)
        else:
            physics_client = pybullet.connect(pybullet.DIRECT)
            threading.Thread(
                target=self._stepSimulation,
                args=[physics_client]).start()

        pybullet.setGravity(0, 0, -9.81, physicsClientId=physics_client)
        return physics_client

    def resetSimulation(self, physics_client):
        """
        Resets the simulated instance corresponding to the physics client id.
        All of the objects loaded in the simulation will be destroyed, but the
        instance will still be running
        """
        self._clearInstance(physics_client)
        pybullet.resetSimulation(physicsClientId=physics_client)

    def stopSimulation(self, physics_client):
        """
        Stops the simulated instance corresponding to the physics_client id

        Parameters:
            physics_client - The id of the simulated instance to be stopped
        """
        self._clearInstance(physics_client)
        pybullet.disconnect(physicsClientId=physics_client)

    def spawnPepper(
            self,
            physics_client,
            translation=[0, 0, 0],
            quaternion=[0, 0, 0, 1],
            spawn_ground_plane=False):
        """
        Loads a Pepper model in the simulation

        Parameters:
            physics_client - The id of the simulated instance in which the
            robot is supposed to be spawned
            translation - List containing 3 elements, the spawning translation
            [x, y, z] in the WORLD frame
            quaternions - List containing 4 elements, the spawning rotation as
            a quaternion [x, y, z, w] in the WORLD frame
            spawn_ground_plane - If True, the pybullet_data ground plane will
            be spawned

        Returns:
            pepper - A PepperVirtual object, the Pepper simulated instance
        """
        pepper = PepperVirtual()

        if spawn_ground_plane:
            pybullet.setAdditionalSearchPath(pybullet_data.getDataPath())
            pybullet.loadMJCF(
                "mjcf/ground_plane.xml",
                physicsClientId=physics_client)

        pepper.loadRobot(
            translation,
            quaternion,
            physicsClientId=physics_client)

        return pepper

    def _clearInstance(self, physics_client):
        """
        INTERNAL METHOD, Called to kill the processes running in a simulated
        instance, before resetting or stopping it. For now, only the robot's
        cameras are cleared

        Parameters:
            physics_client - The client id of the simulated instance that will
            be cleared
        """
        for camera in Camera._getInstances():
            if camera.physics_client == physics_client:
                camera._resetActiveCamera()

    def _stepSimulation(self, physics_client):
        """
        INTERNAL METHOD: This method is only used for a simulation in DIRECT
        mode (without the gui).

        Parameters:
            physics_client - The id of the simulated instance to be stepped
        """
        try:
            while True:
                pybullet.stepSimulation(physicsClientId=physics_client)
                time.sleep(1./240.)
        except Exception:
            pass



#!/usr/bin/env python
# coding: utf-8

import os
import sys
import rospy
import roslib
import roslaunch
import tf
import tf2_ros
import pybullet
from qibullet.camera import Camera
from qibullet.pepper_virtual import PepperVirtual
from threading import Thread
from cv_bridge import CvBridge
from sensor_msgs.msg import Image
from sensor_msgs.msg import CameraInfo
from sensor_msgs.msg import JointState
from std_msgs.msg import Header
from std_msgs.msg import Empty
from naoqi_bridge_msgs.msg import JointAnglesWithSpeed
from naoqi_bridge_msgs.msg import PoseStampedWithSpeed
from geometry_msgs.msg import TransformStamped
from nav_msgs.msg import Odometry

TOP_OPTICAL_FRAME = "CameraTop_optical_frame"
BOTTOM_OPTICAL_FRAME = "CameraBottom_optical_frame"
DEPTH_OPTICAL_FRAME = "CameraDepth_optical_frame"

class PepperRosWrapper:
    """
    Class describing a ROS wrapper for the virtual model of Pepper
    """

    def __init__(self):
        """
        Constructor
        """
        self.spinThread = None
        self.image_bridge = CvBridge()
        self.front_info_msg = dict()
        self.bottom_info_msg = dict()
        self.depth_info_msg = dict()
        self._loadCameraInfos()
        self.transform_broadcaster = tf2_ros.TransformBroadcaster()

    def launchWrapper(self, virtual_pepper, ros_namespace):
        """
        Launches the ROS wrapper for the pepper_virtual instance

        Parameters:
            virtual_pepper - The instance of the simulated model
            ros_namespace - The ROS namespace to be added before the ROS topics
            advertized and subscribed
        """
        self.ros_namespace = ros_namespace
        self.virtual_pepper = virtual_pepper

        rospy.init_node(
            "pybullet_pepper",
            anonymous=True,
            disable_signals=False)

        self.front_cam_pub = rospy.Publisher(
            self.ros_namespace + '/camera/front/image_raw',
            Image,
            queue_size=10)

        self.front_info_pub = rospy.Publisher(
            self.ros_namespace + '/camera/front/camera_info',
            CameraInfo,
            queue_size=10)

        self.bottom_cam_pub = rospy.Publisher(
            self.ros_namespace + '/camera/bottom/image_raw',
            Image,
            queue_size=10)

        self.bottom_info_pub = rospy.Publisher(
            self.ros_namespace + '/camera/bottom/camera_info',
            CameraInfo,
            queue_size=10)

        self.depth_cam_pub = rospy.Publisher(
            self.ros_namespace + '/camera/depth/image_raw',
            Image,
            queue_size=10)

        self.depth_info_pub = rospy.Publisher(
            self.ros_namespace + '/camera/depth/camera_info',
            CameraInfo,
            queue_size=10)

        self.joint_states_pub = rospy.Publisher(
            '/joint_states',
            JointState,
            queue_size=10)

        self.odom_pub = rospy.Publisher(
            'odom',
            Odometry,
            queue_size=10)

        rospy.Subscriber(
            '/joint_angles',
            JointAnglesWithSpeed,
            self.setAnglesCallback)

        rospy.Subscriber(
            '/move_base_simple/goal',
            PoseStampedWithSpeed,
            self.moveToCallback)

        rospy.Subscriber(
            '/move_base_simple/cancel',
            Empty,
            self.killMoveCallback)

        try:
            package_path = roslib.packages.get_pkg_dir("naoqi_driver")
            # path = os.path.dirname(os.path.abspath(__file__)) + "/"

            with open(package_path + "/share/urdf/pepper.urdf", 'r') as file:
                robot_description = file.read()

            # robot_description = robot_description.replace(
            #     "meshes/",
            #     "package://pepper_meshes/meshes/1.0/")

            rospy.set_param("/robot_description", robot_description)

            robot_state_publisher = roslaunch.core.Node(
                "robot_state_publisher",
                "robot_state_publisher")
            roslauncher = roslaunch.scriptapi.ROSLaunch()

            roslauncher.start()
            roslauncher.launch(robot_state_publisher)

            self.spinThread = Thread(target=self.spin)
            self.spinThread.start()

        except IOError as e:
            print("Could not retrieve robot descrition: " + str(e))
            return

    def setUpOdom(self):
        """
        Set up the odometry by broadcasting the tranform and update the
        transform position
        """
        # Send Transform odom
        x, y, theta = self.virtual_pepper.getPosition()
        odom_trans = TransformStamped()
        odom_trans.header.frame_id = "odom"
        odom_trans.child_frame_id = "base_footprint"
        odom_trans.header.stamp = rospy.get_rostime()
        odom_trans.transform.translation.x = x
        odom_trans.transform.translation.y = y
        odom_trans.transform.translation.z = 0
        quaternion = pybullet.getQuaternionFromEuler([0, 0, theta])
        odom_trans.transform.rotation.x = quaternion[0]
        odom_trans.transform.rotation.y = quaternion[1]
        odom_trans.transform.rotation.z = quaternion[2]
        odom_trans.transform.rotation.w = quaternion[3]
        self.transform_broadcaster.sendTransform(odom_trans)
        # Set up the odometry
        odom = Odometry()
        odom.header.stamp = rospy.get_rostime()
        odom.header.frame_id = "odom"
        odom.pose.pose.position.x = x
        odom.pose.pose.position.y = y
        odom.pose.pose.position.z = 0.0
        odom.pose.pose.orientation = odom_trans.transform.rotation
        odom.child_frame_id = "base_footprint"
        [vx, vy, vz], [wx, wy, wz] = pybullet.getBaseVelocity(
            self.virtual_pepper.robot_model,
            self.virtual_pepper.getPhysicsClientId())
        odom.twist.twist.linear.x = vx
        odom.twist.twist.linear.y = vy
        odom.twist.twist.angular.z = wz
        self.odom_pub.publish(odom)

    def getAnglesMsg(self):
        """
        Return the JointState of each joint of the robot
        """
        msg_joint_state = JointState()
        msg_joint_state.header = Header()
        msg_joint_state.header.stamp = rospy.get_rostime()
        msg_joint_state.name = list(self.virtual_pepper.joint_dict)
        msg_joint_state.position = self.virtual_pepper.getAnglesPosition(
            msg_joint_state.name)
        msg_joint_state.name += ["WheelFL", "WheelFR", "WheelB"]
        msg_joint_state.position += [0, 0, 0]
        return msg_joint_state

    def setAnglesCallback(self, msg):
        """
        Callback for setting the angles on a joint list

        Parameters:
            msg - naoqi_bridge_msgs JointAnglesWithSpeed
        """
        joint_list = msg.joint_names
        position_list = list(msg.joint_angles)
        velocity = msg.speed
        self.virtual_pepper.setAngles(joint_list, position_list, velocity)

    def moveToCallback(self, msg):
        """
        Callback to move the robot in frame world or robot

        Parameters:
            msg - naoqi_bridge_msgs PoseStampedWithSpeed
        """
        x = msg.pose_stamped.pose.position.x
        y = msg.pose_stamped.pose.position.y
        theta = pybullet.getEulerFromQuaternion([
            msg.pose_stamped.pose.orientation.x,
            msg.pose_stamped.pose.orientation.y,
            msg.pose_stamped.pose.orientation.z,
            msg.pose_stamped.pose.orientation.w])[-1]

        speed = msg.speed_percentage * self.virtual_pepper.getMaxVelXY()\
            + self.virtual_pepper.getMinVelXY()

        frame = msg.referenceFrame
        self.virtual_pepper.moveTo(x, y, theta, frame=frame, speed=speed)

    def killMoveCallback(self, msg):
        """
        Callback to stop the robot movement
        """
        self.virtual_pepper.moveTo(0, 0, 0)

    def spin(self):
        """
        Overloads the ROS spin method
        """
        try:
            while not rospy.is_shutdown():
                self.joint_states_pub.publish(self.getAnglesMsg())
                self.setUpOdom()
                resolution = self.virtual_pepper.getCameraResolution()
                frame = self.virtual_pepper.getCameraFrame()

                if frame is None or resolution is None:
                    continue

                camera_image_msg = self.image_bridge.cv2_to_imgmsg(frame)

                if self.virtual_pepper.camera_top.isActive():
                    camera_image_msg.encoding = "bgr8"
                    camera_image_msg.header.frame_id = TOP_OPTICAL_FRAME

                    if resolution == Camera.K_VGA:
                        camera_info_msg = self.front_info_msg["K_VGA"]
                    elif resolution == Camera.K_QVGA:
                        camera_info_msg = self.front_info_msg["K_QVGA"]
                    elif resolution == Camera.K_QQVGA:
                        camera_info_msg = self.front_info_msg["K_QQVGA"]

                    self.front_cam_pub.publish(camera_image_msg)
                    self.front_info_pub.publish(camera_info_msg)

                elif self.virtual_pepper.camera_bottom.isActive():
                    camera_image_msg.encoding = "bgr8"
                    camera_image_msg.header.frame_id = BOTTOM_OPTICAL_FRAME

                    if resolution == Camera.K_VGA:
                        camera_info_msg = self.bottom_info_msg["K_VGA"]
                    elif resolution == Camera.K_QVGA:
                        camera_info_msg = self.bottom_info_msg["K_QVGA"]
                    elif resolution == Camera.K_QQVGA:
                        camera_info_msg = self.bottom_info_msg["K_QQVGA"]

                    self.bottom_cam_pub.publish(camera_image_msg)
                    self.bottom_info_pub.publish(camera_info_msg)

                elif self.virtual_pepper.camera_depth.isActive():
                    camera_image_msg.encoding = "mono16"
                    camera_image_msg.header.frame_id = DEPTH_OPTICAL_FRAME

                    if resolution == Camera.K_VGA:
                        camera_info_msg = self.depth_info_msg["K_VGA"]
                    elif resolution == Camera.K_QVGA:
                        camera_info_msg = self.depth_info_msg["K_QVGA"]
                    elif resolution == Camera.K_QQVGA:
                        camera_info_msg = self.depth_info_msg["K_QQVGA"]

                    self.depth_cam_pub.publish(camera_image_msg)
                    self.depth_info_pub.publish(camera_info_msg)

        except Exception as e:
            print("Stopping the ROS wrapper")

    def _loadCameraInfos(self):
        """
        INTERNAL METHOD, creates the camera info message for Pepper's cameras
        """
        for quality in ["K_VGA", "K_QVGA", "K_QQVGA"]:
            self.front_info_msg[quality] = CameraInfo()
            self.front_info_msg[quality].header.frame_id =\
                "CameraTop_optical_frame"
            self.front_info_msg[quality].distortion_model = "plumb_bob"

            self.bottom_info_msg[quality] = CameraInfo()
            self.bottom_info_msg[quality].header.frame_id =\
                "CameraBottom_optical_frame"
            self.bottom_info_msg[quality].distortion_model = "plumb_bob"

            self.depth_info_msg[quality] = CameraInfo()
            self.depth_info_msg[quality].header.frame_id =\
                "CameraDepth_optical_frame"
            self.depth_info_msg[quality].distortion_model = "plumb_bob"

        self.front_info_msg["K_VGA"].width = Camera.K_VGA.width
        self.front_info_msg["K_VGA"].height = Camera.K_VGA.height
        self.front_info_msg["K_VGA"].D = [
            -0.0545211535376379,
            0.0691973423510287,
            -0.00241094929163055,
            -0.00112245009306511,
            0]
        self.front_info_msg["K_VGA"].K = [
            556.845054830986, 0, 309.366895338178,
            0, 555.898679730161, 230.592233628776,
            0, 0, 1]
        self.front_info_msg["K_VGA"].R = [
            1, 0, 0,
            0, 1, 0,
            0, 0, 1]
        self.front_info_msg["K_VGA"].P = [
            551.589721679688, 0, 308.271132841983, 0,
            0, 550.291320800781, 229.20143668168, 0,
            0, 0, 1, 0]

        self.front_info_msg["K_QVGA"].width = Camera.K_QVGA.width
        self.front_info_msg["K_QVGA"].height = Camera.K_QVGA.height
        self.front_info_msg["K_QVGA"].D = [
            -0.0870160932911717,
            0.128210165050533,
            0.003379500659424,
            -0.00106205540818586,
            0]
        self.front_info_msg["K_QVGA"].K = [
            274.139508945831, 0, 141.184472810944,
            0, 275.741846757374, 106.693773654172,
            0, 0, 1]
        self.front_info_msg["K_QVGA"].R = [
            1, 0, 0,
            0, 1, 0,
            0, 0, 1]
        self.front_info_msg["K_QVGA"].P = [
            272.423675537109, 0, 141.131930791285, 0,
            0, 273.515747070312, 107.391746054313, 0,
            0, 0, 1, 0]

        self.front_info_msg["K_QQVGA"].width = Camera.K_QQVGA.width
        self.front_info_msg["K_QQVGA"].height = Camera.K_QQVGA.height
        self.front_info_msg["K_QQVGA"].D = [
            -0.0843564504845967,
            0.125733083790192,
            0.00275901756247071,
            -0.00138645823460527,
            0]
        self.front_info_msg["K_QQVGA"].K = [
            139.424539568966, 0, 76.9073669920582,
            0, 139.25542782325, 59.5554242026743,
            0, 0, 1]
        self.front_info_msg["K_QQVGA"].R = [
            1, 0, 0,
            0, 1, 0,
            0, 0, 1]
        self.front_info_msg["K_QQVGA"].P = [
            137.541534423828, 0, 76.3004646597892, 0,
            0, 136.815216064453, 59.3909799751191, 0,
            0, 0, 1, 0]

        self.bottom_info_msg["K_VGA"].width = Camera.K_VGA.width
        self.bottom_info_msg["K_VGA"].height = Camera.K_VGA.height
        self.bottom_info_msg["K_VGA"].D = [
            -0.0648763971625288,
            0.0612520196884308,
            0.0038281538281731,
            -0.00551104078371959,
            0]
        self.bottom_info_msg["K_VGA"].K = [
            558.570339530768, 0, 308.885375457296,
            0, 556.122943034837, 247.600724811385,
            0, 0, 1]
        self.bottom_info_msg["K_VGA"].R = [
            1, 0, 0,
            0, 1, 0,
            0, 0, 1]
        self.bottom_info_msg["K_VGA"].P = [
            549.571655273438, 0, 304.799679526441, 0,
            0, 549.687316894531, 248.526959297022, 0,
            0, 0, 1, 0]

        self.bottom_info_msg["K_QVGA"].width = Camera.K_QVGA.width
        self.bottom_info_msg["K_QVGA"].height = Camera.K_QVGA.height
        self.bottom_info_msg["K_QVGA"].D = [
            -0.0481869853715082,
            0.0201858398559121,
            0.0030362056699177,
            -0.00172241952442813,
            0]
        self.bottom_info_msg["K_QVGA"].K = [
            278.236008818534, 0, 156.194471689706,
            0, 279.380102992049, 126.007123836447,
            0, 0, 1]
        self.bottom_info_msg["K_QVGA"].R = [
            1, 0, 0,
            0, 1, 0,
            0, 0, 1]
        self.bottom_info_msg["K_QVGA"].P = [
            273.491455078125, 0, 155.112454709117, 0,
            0, 275.743133544922, 126.057357467223, 0,
            0, 0, 1, 0]

        self.bottom_info_msg["K_QQVGA"].width = Camera.K_QQVGA.width
        self.bottom_info_msg["K_QQVGA"].height = Camera.K_QQVGA.height
        self.bottom_info_msg["K_QQVGA"].D = [
            -0.0688388724945936,
            0.0697453843669642,
            0.00309518737071049,
            -0.00570486993696543,
            0]
        self.bottom_info_msg["K_QQVGA"].K = [
            141.611855886672, 0, 78.6494086288656,
            0, 141.367163830175, 58.9220646201529,
            0, 0, 1]
        self.bottom_info_msg["K_QQVGA"].R = [
            1, 0, 0,
            0, 1, 0,
            0, 0, 1]
        self.bottom_info_msg["K_QQVGA"].P = [
            138.705535888672, 0, 77.2544255212306, 0,
            0, 138.954086303711, 58.7000861760043, 0,
            0, 0, 1, 0]

        self.depth_info_msg["K_VGA"].width = Camera.K_VGA.width
        self.depth_info_msg["K_VGA"].height = Camera.K_VGA.height
        self.depth_info_msg["K_VGA"].D = [
            -0.0688388724945936,
            0.0697453843669642,
            0.00309518737071049,
            -0.00570486993696543,
            0]
        self.bottom_info_msg["K_VGA"].K = [
            525, 0, 319.5000000,
            0, 525, 239.5000000000000,
            0, 0, 1]
        self.bottom_info_msg["K_VGA"].R = [
            1, 0, 0,
            0, 1, 0,
            0, 0, 1]
        self.bottom_info_msg["K_VGA"].P = [
            525, 0, 319.500000, 0,
            0, 525, 239.5000000000, 0,
            0, 0, 1, 0]

        self.depth_info_msg["K_QVGA"].width = Camera.K_QVGA.width
        self.depth_info_msg["K_QVGA"].height = Camera.K_QVGA.height
        self.depth_info_msg["K_QVGA"].D = [
            -0.0688388724945936,
            0.0697453843669642,
            0.00309518737071049,
            -0.00570486993696543,
            0]
        self.bottom_info_msg["K_QVGA"].K = [
            525/2.0, 0, 319.5000000/2.0,
            0, 525/2.0, 239.5000000000000/2.0,
            0, 0, 1]
        self.bottom_info_msg["K_QVGA"].R = [
            1, 0, 0,
            0, 1, 0,
            0, 0, 1]
        self.bottom_info_msg["K_QVGA"].P = [
            525/2.0, 0, 319.500000/2.0, 0,
            0, 525/2.0, 239.5000000000/2.0, 0,
            0, 0, 1, 0]

        self.depth_info_msg["K_QQVGA"].width = Camera.K_QQVGA.width
        self.depth_info_msg["K_QQVGA"].height = Camera.K_QQVGA.height
        self.depth_info_msg["K_QQVGA"].D = [
            -0.0688388724945936,
            0.0697453843669642,
            0.00309518737071049,
            -0.00570486993696543,
            0]
        self.bottom_info_msg["K_QQVGA"].K = [
            525/4.0, 0, 319.5000000/4.0,
            0, 525/4.0, 239.5000000000000/4.0,
            0, 0, 1]
        self.bottom_info_msg["K_QQVGA"].R = [
            1, 0, 0,
            0, 1, 0,
            0, 0, 1]
        self.bottom_info_msg["K_QQVGA"].P = [
            525/4.0, 0, 319.500000/4.0, 0,
            0, 525/4.0, 239.5000000000/4.0, 0,
            0, 0, 1, 0]



#!/usr/bin/env python
# coding: utf-8

import sys
import pybullet
from qibullet.link import Link
from qibullet.joint import Joint

IS_VERSION_PYTHON_3 = sys.version_info[0] >= 3


class RobotVirtual:
    """
    Mother class representing a virtual robot
    """

    def __init__(self, description_file):
        """
        Constructor

        Parameters:
            - description_file: The file giving the description of the virtual
            robot. For now, only URDF is handled
        """
        self.description_file = description_file
        self.physics_client = 0
        self.joint_dict = dict()
        self.link_dict = dict()

    def loadRobot(self, translation, quaternion, physicsClientId=0):
        """
        Loads the robot into a simulation, loads the joints and the links
        descriptions. The joints are set to 0 rad

        Parameters:
            translation - List containing 3 elements, the translation [x, y, z]
            of the robot in the WORLD frame
            quaternion - List containing 4 elements, the quaternion
            [x, y, z, q] of the robot in the WORLD frame
            physicsClientId - The id of the simulated instance in which the
            robot is supposed to be loaded

        Returns:
            boolean - True if the method ran correctly, False otherwise
        """
        try:
            self.physics_client = physicsClientId
            self.robot_model = pybullet.loadURDF(
                self.description_file,
                translation,
                quaternion,
                useFixedBase=False,
                globalScaling=1.0,
                physicsClientId=self.physics_client,
                flags=pybullet.URDF_USE_SELF_COLLISION |
                pybullet.URDF_USE_MATERIAL_COLORS_FROM_MTL)

        except pybullet.error as e:
            print("Cannot load robot model: " + str(e))
            return False

        for i in range(pybullet.getNumJoints(
                self.robot_model,
                physicsClientId=self.physics_client)):
            if IS_VERSION_PYTHON_3:
                # PYTHON 3 version needs a conversion bytes to str
                joint_info = pybullet.getJointInfo(
                    self.robot_model,
                    i,
                    physicsClientId=self.physics_client)
                self.link_dict[joint_info[12].decode('utf-8')] =\
                    Link(joint_info)

                if joint_info[2] == pybullet.JOINT_PRISMATIC or\
                        joint_info[2] == pybullet.JOINT_REVOLUTE:
                    self.joint_dict[joint_info[1].decode('utf-8')] =\
                        Joint(joint_info)
            else:
                # PYTHON 2 Version
                joint_info = pybullet.getJointInfo(
                    self.robot_model,
                    i,
                    physicsClientId=self.physics_client)

                self.link_dict[joint_info[12]] = Link(joint_info)

                if joint_info[2] == pybullet.JOINT_PRISMATIC or\
                        joint_info[2] == pybullet.JOINT_REVOLUTE:
                    self.joint_dict[joint_info[1]] = Joint(joint_info)

            pybullet.setJointMotorControl2(
                self.robot_model,
                i,
                pybullet.POSITION_CONTROL,
                0,
                physicsClientId=self.physics_client)

        return True

    def getPhysicsClientId(self):
        """
        Gets the id of the id of the simulation in which this robot instance is
        loaded

        Returns:
            physics_client - The id of the simulation
        """
        return self.physics_client

    def setAngles(self, joint_names, joint_values, percentage_speed):
        """
        Set angles on the robot's joints. Tests have to be performed by the
        child class to guarantee the validity of the input parameters

        Parameters:
            joint_names - List of string containing the name of the joints
            to be controlled
            joint_values - List of values corresponding to the angles in
            radians to be applied
            percentage_speed - Percentage of the max speed to be used for the
            movement, has to be strictly superior to 0 and inferior or equal to
            1
        """
        try:
            assert len(joint_names) == len(joint_values)
            assert percentage_speed > 0.0 and percentage_speed <= 1.0

        except AssertionError:
            print("Error in the setAngles parameters")
            return

        joint_speed = 0

        for joint_name, joint_value in zip(joint_names, joint_values):
            joint_speed =\
                self.joint_dict[joint_name].getMaxVelocity() *\
                percentage_speed

            pybullet.setJointMotorControl2(
                self.robot_model,
                self.joint_dict[joint_name].getIndex(),
                pybullet.POSITION_CONTROL,
                targetPosition=joint_value,
                maxVelocity=joint_speed,
                physicsClientId=self.physics_client)

    def getAnglesPosition(self, joint_names):
        """
        Gets the position of the robot's joints in radians. If one of the joint
        doesn't exist, the method will raise a KeyError

        Parameters:
            joint_names - List of string containing the names of the joints

        Returns:
            joint_positions - List of floats containing the joint's positions
        """
        joint_positions = list()

        for joint_name in joint_names:
            joint_positions.append(pybullet.getJointState(
                self.robot_model,
                self.joint_dict[joint_name].getIndex(),
                physicsClientId=self.physics_client)[0])

        return joint_positions

    def getPosition(self):
        """
        Gets the position of the robot's base in the world frame

        Returns:
            x - The position of the robot's base on the x axis, in meters
            y - The positions of the robot's base on the y axis in meters
            theta - The rotation of the robot's base on the z axis in meters
        """
        position, quaternions = pybullet.getBasePositionAndOrientation(
            self.robot_model,
            physicsClientId=self.physics_client)

        theta = pybullet.getEulerFromQuaternion(quaternions)[2]
        return position[0], position[1], theta



#!/usr/bin/env python
# coding: utf-8

import os
import time
import pybullet
from qibullet.tools import *
from qibullet.camera import *
from qibullet.robot_virtual import RobotVirtual

MAX_VEL_XY = 0.55
MIN_VEL_XY = 0.1
MAX_ACC_XY = 0.55
MIN_ACC_XY = 0.3
MAX_VEL_THETA = 2.0
MIN_VEL_THETA = 0.3
MAX_ACC_THETA = 3.0
MIN_ACC_THETA = 0.75


class PepperVirtual(RobotVirtual):
    """
    Class representing the virtual instance of the Pepper robot
    """
    ID_CAMERA_TOP = 0
    ID_CAMERA_BOTTOM = 1
    ID_CAMERA_DEPTH = 2
    FRAME_WORLD = 1
    FRAME_ROBOT = 2
    URDF_PATH = "robot_data/pepper_1.7/pepper.urdf"

    def __init__(self):
        """
        Constructor
        """
        RobotVirtual.__init__(self, PepperVirtual.URDF_PATH)
        self.camera_top = None
        self.camera_bottom = None
        self.camera_depth = None
        self.motion_constraint = None
        # Default speed (in m/s) xy : 0.35, min : 0.1, max : 0.55
        self.vel_xy = 0.35
        # Default acc (in m/s^2 xy : 0.3, min : 0.1, max : 0.55
        self.acc_xy = 0.3
        # Default speed (in rad/s) theta : 1.0, min : 0.2, max : 2.0
        self.vel_theta = 1.0
        # Default acc (in rad/s^2 theta : 0.75, min : 0.1, max : 3.0
        self.acc_theta = 0.3

    def loadRobot(self, translation, quaternion, physicsClientId=0):
        """
        Overload @loadRobot from the @RobotVirtual class. Update max velocity
        for the fingers and thumbs, based on the hand joints. Add self
        collision exceptions (The biceps won't autocollide with the torso, the
        fingers and thumbs of a hand won't autocollide with the corresponding
        wrist). Add the cameras. Add motion constraint.
        """
        pybullet.setAdditionalSearchPath(
            os.path.dirname(os.path.realpath(__file__)),
            physicsClientId=physicsClientId)

        RobotVirtual.loadRobot(
            self,
            translation,
            quaternion,
            physicsClientId=physicsClientId)

        for joint_name in list(self.joint_dict):
            if 'RFinger' in joint_name or 'RThumb' in joint_name:
                self.joint_dict[joint_name].setMaxVelocity(
                    self.joint_dict["RHand"].getMaxVelocity())
            elif 'LFinger' in joint_name or 'LThumb' in joint_name:
                self.joint_dict[joint_name].setMaxVelocity(
                    self.joint_dict["LHand"].getMaxVelocity())
            elif "Wheel" in joint_name:
                    self.joint_dict.pop(joint_name)

        for shoulder_roll_link in ["RBicep", "LBicep"]:
            pybullet.setCollisionFilterPair(
                self.robot_model,
                self.robot_model,
                self.link_dict["torso"].getIndex(),
                self.link_dict[shoulder_roll_link].getIndex(),
                0,
                physicsClientId=self.physics_client)

        for name, link in self.link_dict.items():
            for wrist in ["r_wrist", "l_wrist"]:
                if wrist[0] + "finger" in name.lower() or\
                   wrist[0] + "thumb" in name.lower():
                    pybullet.setCollisionFilterPair(
                        self.robot_model,
                        self.robot_model,
                        self.link_dict[wrist].getIndex(),
                        link.getIndex(),
                        0,
                        physicsClientId=self.physics_client)

        self.camera_top = CameraRgb(
            self.robot_model,
            self.link_dict["CameraTop_optical_frame"],
            physicsClientId=self.physics_client)

        self.camera_bottom = CameraRgb(
            self.robot_model,
            self.link_dict["CameraBottom_optical_frame"],
            physicsClientId=self.physics_client)

        self.camera_depth = CameraDepth(
            self.robot_model,
            self.link_dict["CameraDepth_optical_frame"],
            physicsClientId=self.physics_client)

        # TODO: motion constraint fucks up the right hand mimic behaviour
        self.motion_constraint = pybullet.createConstraint(
            parentBodyUniqueId=self.robot_model,
            parentLinkIndex=-1,
            childBodyUniqueId=-1,
            childLinkIndex=-1,
            jointType=pybullet.JOINT_FIXED,
            jointAxis=[0, 0, 0],
            parentFramePosition=[0, 0, 0],
            parentFrameOrientation=[0, 0, 0, 1],
            childFramePosition=[translation[0], translation[1], 0],
            childFrameOrientation=quaternion,
            physicsClientId=self.physics_client)

    def moveTo(self, x, y, theta, frame=FRAME_ROBOT, speed=None):
        """
        Move the robot in frame world or robot.
        (FRAME_WORLD = 1, FRAME_ROBOT = 2)

        Parameters:
            x - float in meters
            y - float in meters
            theta - float in radians
        """

        # force applied in the movement
        force = 100
        # The robot will stop the movement with a precision
        # of 0.01 m and 0.02 rads
        threshold_xy = 0.01
        threshold_theta = 0.02

        # get actual position in frame world
        actual_pose, actual_orn = pybullet.getBasePositionAndOrientation(
            self.robot_model,
            physicsClientId=self.physics_client)
        # pose x, y, z
        pose_requested = [x, y, 0]
        # orientation requested (quaternions)
        orn_requested = pybullet.getQuaternionFromEuler([0, 0, theta])
        # if we are in frame robot add position in the frame world
        if frame == 2:
            orn_euler = pybullet.getEulerFromQuaternion(actual_orn)
            pose_requested = [
                pose_requested[0] * math.cos(orn_euler[2])
                - pose_requested[1] * math.sin(orn_euler[2])
                + actual_pose[0],
                pose_requested[0] * math.sin(orn_euler[2])
                + pose_requested[1] * math.cos(orn_euler[2])
                + actual_pose[1],
                0]
            orn_requested = pybullet.getQuaternionFromEuler(
                [orn_euler[0],
                 orn_euler[1],
                 orn_euler[2] + theta])
        # change the constraint to the position and orientation requested
        pybullet.changeConstraint(
            self.motion_constraint,
            pose_requested,
            jointChildFrameOrientation=orn_requested,
            maxForce=force,
            physicsClientId=self.physics_client)
        # init robot speed
        speed_xy = self.vel_xy
        if speed is not None:
            speed_xy = speed
        vel_x, vel_y, vel_theta = [speed_xy, speed_xy, self.vel_theta]
        # Compute the ratio distance requested on distance total
        distance = getDistance(actual_pose, pose_requested)
        p_x = 0
        p_y = 0
        p_theta = 0
        if distance:
            p_x = (pose_requested[0] - actual_pose[0]) / distance
            p_y = (pose_requested[1] - actual_pose[1]) / distance
        theta_to_do = getOrientation(actual_orn, orn_requested)
        if abs(theta_to_do):
            p_theta = abs(theta_to_do) / theta_to_do

        pose_init = actual_pose
        orn_init = actual_orn
        while getDistance(actual_pose, pose_requested) > threshold_xy\
                or abs(getOrientation(actual_orn, orn_requested)) >\
                threshold_theta:

            actual_pose, actual_orn = pybullet.getBasePositionAndOrientation(
                self.robot_model,
                physicsClientId=self.physics_client)
            vel_x = computeVelocity(
                        self.acc_xy,
                        0.05,
                        speed_xy,
                        getDistance(pose_init, actual_pose),
                        getDistance(actual_pose, pose_requested)
                        )
            vel_y = vel_x
            vel_theta = computeVelocity(
                        self.acc_theta,
                        0.05,
                        self.vel_theta,
                        abs(getOrientation(orn_init, orn_requested)),
                        abs(getOrientation(actual_orn, orn_requested))
                        )
            # if the robot is on the position requested, we set the
            # velocity to 0.
            if abs(actual_pose[0] - pose_requested[0]) <= threshold_xy / 2:
                vel_x = 0
            if abs(actual_pose[1] - pose_requested[1]) <= threshold_xy / 2:
                vel_y = 0
            if abs(getOrientation(actual_orn, orn_requested)) <=\
                    threshold_theta:
                vel_theta = 0
            # reset velocity of the robot
            time.sleep(0.02)
            pybullet.resetBaseVelocity(
                self.robot_model,
                [vel_x * p_x, vel_y * p_y, 0],
                [0, 0, vel_theta * p_theta],
                physicsClientId=self.physics_client)
        # Change the constraint to the actual position and orientation in
        # order to stop the robot's motion. The force applied is huge
        # to avoid oscillation.
        actual_pose, actual_orn = pybullet.getBasePositionAndOrientation(
            self.robot_model,
            physicsClientId=self.physics_client)
        pybullet.changeConstraint(
            self.motion_constraint,
            actual_pose,
            jointChildFrameOrientation=actual_orn,
            maxForce=force * 10,
            physicsClientId=self.physics_client)
        pybullet.resetBaseVelocity(
            self.robot_model,
            [0, 0, 0],
            [0, 0, 0],
            physicsClientId=self.physics_client)

    def setAngles(self, joint_names, joint_values, percentage_speed):
        """
        Overload @setAngles from the @RobotVirtual class. Handles the finger
        mimic behaviour.

        Parameters:
            joint_names - List of string (or string if only one joint)
            containing the name of the joints to be controlled
            joint_values - List of values (or value if only one joint)
            corresponding to the angles in radians to be applied
            percentage_speed - Percentage of the max speed to be used for the
            movement
        """
        try:
            if type(joint_names) is str:
                assert type(joint_values) is int or type(joint_values) is float
                names = [joint_names]
                values = [joint_values]
            else:
                assert type(joint_names) is type(joint_values) is list
                names = list(joint_names)
                values = list(joint_values)

        except AssertionError:
            print("Error in the parameters given to the setAngles method")
            return

        for hand in ["RHand", "LHand"]:
            for i in range(names.count(hand)):
                index = names.index(hand)
                value = values[index]
                names.pop(index)
                values.pop(index)
                finger_names, finger_values = self._mimicHand(hand, value)
                names.extend(finger_names)
                values.extend(finger_values)

        RobotVirtual.setAngles(
            self,
            names,
            values,
            percentage_speed)

    def getAnglesPosition(self, joint_names):
        """
        Overload @getAnglesPosition from the @RobotVirtual class. Handles the
        finger mimicked position for the hands (when getting the position of
        RHand or LHand, will return the hand's opening percentage).

        Parameters:
            joint_names - List of string (or string if only one joint)
            containing the name of the joints

        Returns:
            joint_positions - List of float (or float if only one joint)
            containing the joint's positions in radians
        """
        if type(joint_names) is str:
            names = [joint_names]
        else:
            names = list(joint_names)

        joint_positions = RobotVirtual.getAnglesPosition(self, names)

        for hand, finger in zip(
                ["RHand", "LHand"],
                ["RFinger11", "LFinger11"]):
            for i in range(names.count(hand)):
                index = names.index(hand)
                joint_positions[index] =\
                    RobotVirtual.getAnglesPosition(self, [finger]).pop() /\
                    (1/(self.joint_dict[finger].getUpperLimit() -
                        self.joint_dict[finger].getLowerLimit())) +\
                    self.joint_dict[finger].getLowerLimit()

        if len(joint_positions) == 1:
            return joint_positions.pop()
        else:
            return joint_positions

    def setVelXY(self, value):
        """
        Set xy velocity of the robot

        Parameter:
            value of the velocity in m/s
        """
        if value >= MAX_VEL_XY:
            self.vel_xy = MAX_VEL_XY
        elif value <= MIN_VEL_XY:
            self.vel_xy = MIN_VEL_XY
        else:
            self.vel_xy = value

    def setVelTheta(self, value):
        """
        Set theta velocity of the robot

        Parameter:
            value of the velocity in m/s
        """
        if value >= MAX_VEL_THETA:
            self.vel_theta = MAX_VEL_THETA
        elif value <= MIN_VEL_THETA:
            self.vel_theta = MIN_VEL_THETA
        else:
            self.vel_theta = value

    def setAccXY(self, value):
        """
        Set xy acceleration of the robot

        Parameter:
            value of the velocity in m/s^2
        """
        if value >= MAX_ACC_XY:
            self.acc_xy = MAX_ACC_XY
        elif value <= MIN_ACC_XY:
            self.acc_xy = MIN_ACC_XY
        else:
            self.acc_xy = value

    def setAccTheta(self, value):
        """
        Set theta acceleration of the robot

        Parameter:
            value of the velocity in rad/s^2
        """
        if value >= MAX_ACC_THETA:
            self.acc_theta = MAX_ACC_THETA
        elif value <= MIN_ACC_THETA:
            self.acc_theta = MIN_ACC_THETA
        else:
            self.acc_theta = value

    def getMaxVelXY(self):
        return MAX_VEL_XY

    def getMinVelXY(self):
        return MIN_VEL_XY

    def getMaxVelTheta(self):
        return MAX_VEL_THETA

    def getMinVelTheta(self):
        return MIN_VEL_THETA

    def subscribeCamera(self, camera_id, resolution=Camera.K_QVGA):
        """
        Subscribe to the camera holding the camera id. WARNING: at the moment,
        only one camera can be subscribed.

        Parameters:
            camera_id - The id of the camera to be subscribed
            resolution - CameraResolution object, the resolution of the camera
        """
        if camera_id == PepperVirtual.ID_CAMERA_TOP:
            self.camera_top.subscribe(resolution=resolution)

        elif camera_id == PepperVirtual.ID_CAMERA_BOTTOM:
            self.camera_bottom.subscribe(resolution=resolution)

        elif camera_id == PepperVirtual.ID_CAMERA_DEPTH:
            self.camera_depth.subscribe(resolution=resolution)

    def unsubscribeCamera(self, camera_id):
        """
        Unsubscribe from a camera, the one holding the camera id.

        Parameters:
            camera_id - The id of the camera to be unsubscribed
        """
        if camera_id == PepperVirtual.ID_CAMERA_TOP:
            self.camera_top.unsubscribe()

        elif camera_id == PepperVirtual.ID_CAMERA_BOTTOM:
            self.camera_bottom.unsubscribe()

        elif camera_id == PepperVirtual.ID_CAMERA_DEPTH:
            self.camera_depth.unsubscribe()

    def getCameraFrame(self):
        """
        Returns a camera frame. Be advised that the subscribeCamera method
        needs to be called beforehand.

        Returns:
            frame - The current camera frame as a formatted numpy array,
            directly exploitable from OpenCV
        """
        if self.camera_top.isActive():
            return self.camera_top.getFrame()
        elif self.camera_bottom.isActive():
            return self.camera_bottom.getFrame()
        elif self.camera_depth.isActive():
            return self.camera_depth.getFrame()

    def getCameraResolution(self):
        """
        Returns the resolution of the active camera. If no camera is active,
        returns None

        Returns:
            resolution - a CameraResolution object describing the resolution of
            the active camera
        """
        if self.camera_top.isActive():
            return self.camera_top.getResolution()
        elif self.camera_bottom.isActive():
            return self.camera_bottom.getResolution()
        elif self.camera_depth.isActive():
            return self.camera_depth.getResolution()

    def isSelfColliding(self, link_names):
        """
        Specifies if a link is colliding with the rest of the virtual Pepper
        robot.

        Parameters:
            link_names - String or list of string containing the names of the
            links to be checked for self collision. WARNING: only the links
            with corresponding meshes should be used, otherwise the link cannot
            self collide

        Returns:
            self_colliding - Boolean, if True at least one of the links is self
            colliding
        """
        try:
            if type(link_names) is str:
                assert link_names in self.link_dict.keys()
                names = [link_names]
            else:
                assert set(link_names).issubset(self.link_dict.keys())
                names = list(link_names)

            for name in names:
                contact_tuple = pybullet.getContactPoints(
                    bodyA=self.robot_model,
                    bodyB=self.robot_model,
                    linkIndexA=self.link_dict[name].getIndex(),
                    physicsClientId=self.physics_client)
                contact_tuple += pybullet.getContactPoints(
                    bodyA=self.robot_model,
                    bodyB=self.robot_model,
                    linkIndexB=self.link_dict[name].getIndex(),
                    physicsClientId=self.physics_client)

                if len(contact_tuple) != 0:
                    return True

            return False

        except AssertionError:
            "Unauthorized link checking for self collisions"
            return False

    def _mimicHand(self, hand, value, multiplier=0.872665, offset=0):
        """
        Used to propagate a joint value on the fingers attached to the hand.
        The formula used to mimic a joint is the following (with a multiplier
        of 0.872665 and an offset of 0):

        finger_value = (hand_value * multiplier) + offset

        Parameters:
            hand - String, RHand or LHand
            value - the joint value to be propagated

        Returns:
            finger_names - Names of the finger to be controlled
            finger_values - Values of the fingers to be controlled
        """
        finger_names = list()
        finger_values = list()

        for joint_name in self.joint_dict.keys():
            if (hand[0] + "Finger") in joint_name or\
               (hand[0] + "Thumb") in joint_name:
                finger_names.append(joint_name)
                finger_values.append((value * multiplier) + offset)

        return finger_names, finger_values



#!/usr/bin/env python
# coding: utf-8

import atexit
import weakref
import pybullet
import threading
import numpy as np
from qibullet.link import Link


class CameraResolution:
    """
    Enumeration of the camera resolutions
    """

    def __init__(self, width, height):
        """
        Constructor

        Parameters:
            width - Width resolution in pixels
            height - Height resolution in pixels
        """
        self.width = width
        self.height = height

    def __eq__(self, resolution):
        """
        Overloading the equal operator

        Parameters:
            resolution - the comparing resolution
        """
        if self.width == resolution.width and self.height == resolution.height:
            return True
        else:
            return False


class Camera:
    """
    Class representing a virtual camera
    """
    _instances = set()
    ACTIVE_CAMERA_ID = dict()

    K_QQVGA = CameraResolution(160, 120)
    K_QVGA = CameraResolution(320, 240)
    K_VGA = CameraResolution(640, 480)
    K_QQ720p = CameraResolution(320, 180)
    K_Q720p = CameraResolution(640, 360)
    K_720p = CameraResolution(1280, 720)

    def __init__(
            self,
            robot_model,
            link,
            physicsClientId=0,
            near_plane=0.01,
            far_plane=100):
        """
        Constructor

        Parameters:
            robot_model - The pybullet model of the robot
            link - The link (Link type) onto which the camera's attached
            physicsClientId - The id of the simulated instance in which the
            camera is to be spawned
            near_plane - The near plane distance
            far_plane - The far plane distance
        """
        self.robot_model = robot_model
        self.physics_client = physicsClientId
        self.link = link
        self.near_plane = near_plane
        self.far_plane = far_plane
        self.frame = None
        self.projection_matrix = None
        self.resolution = None
        self.fov = None
        self.extraction_thread = threading.Thread()
        self.frame_lock = threading.Lock()
        self.resolution_lock = threading.Lock()
        self._instances.add(weakref.ref(self))
        self._resetActiveCamera()
        atexit.register(self._resetActiveCamera)

    @classmethod
    def _getInstances(cls):
        """
        INTERNAL CLASSMETHOD, get all of the Camera (and daughters) instances
        """
        dead = set()

        for ref in cls._instances:
            obj = ref()

            if obj is not None:
                yield obj
            else:
                dead.add(ref)

        cls._instances -= dead

    def subscribe(self, id, resolution):
        """
        Subscribing method for the camera. The FOV has to be specified
        beforehand

        Parameters:
            id - The id of the Python object calling the method
            resolution - CameraResolution object, the resolution of the camera
        """
        # Lets the extracting thread die before launching another one if the
        # same camera calls the subscribing method for a second time
        self._resetActiveCamera()
        self._setResolution(resolution)
        Camera.ACTIVE_CAMERA_ID[self.physics_client] = id

    def unsubscribe(self, id):
        """
        Method stopping the frame retreival thread for a camera
        """
        if Camera.ACTIVE_CAMERA_ID[self.physics_client] == id:
            self._resetActiveCamera()

    def getFrame(self):
        """
        Returns the current frame

        Returns:
            frame - The current frame
        """
        with self.frame_lock:
            return self.frame

    def isActive(self):
        """
        Specifies if the camera is active or not

        Returns:
            is_active - Boolean, True if the camera is subscribed to, False
            otherwise
        """
        return id(self) == Camera.ACTIVE_CAMERA_ID[self.physics_client]

    def getResolution(self):
        """
        Returns the resolution of the camera

        Returns:
            resolution - A CameraResolution object, the resolution of the
            camera
        """
        with self.resolution_lock:
            return self.resolution

    def _setFov(self, fov):
        """
        INTERNAL METHOD, sets the field of view of the camera

        Parameters:
            fov - The value of the field of view
        """
        try:
            assert type(fov) is int or type(fov) is float
            self.fov = fov

        except AssertionError as e:
            print("Cannot set the camera FOV: " + str(e))

    def _setResolution(self, resolution):
        """
        INTERNAL METHOD, sets the resolution for the camera. The fov of the
        camera has to be setted beforehand

        Parameters:
            resolution - The resolution of the camera (CameraResolution type)
        """
        try:
            with self.resolution_lock:
                assert isinstance(resolution, CameraResolution)
                assert self.fov is not None

                self.resolution = resolution
                self.projection_matrix = pybullet.computeProjectionMatrixFOV(
                    self.fov,
                    self.resolution.width / self.resolution.height,
                    self.near_plane,
                    self.far_plane,
                    physicsClientId=self.physics_client)

        except AssertionError as e:
            print("Cannot set camera resolution: " + str(e))

    def _getCameraImage(self):
        """
        INTERNAL METHOD, Computes the OpenGL virtual camera image. The
        resolution and the projection matrix have to be computed before calling
        this method, or it will crash

        Returns:
            camera_image - The camera image of the OpenGL virtual camera
        """
        _, _, _, _, pos_world, q_world = pybullet.getLinkState(
            self.robot_model,
            self.link.getParentIndex(),
            computeForwardKinematics=True,
            physicsClientId=self.physics_client)

        rotation = pybullet.getMatrixFromQuaternion(q_world)
        forward_vector = [rotation[0], rotation[3], rotation[6]]
        up_vector = [rotation[2], rotation[5], rotation[8]]

        camera_target = [
            pos_world[0] + forward_vector[0] * 10,
            pos_world[1] + forward_vector[1] * 10,
            pos_world[2] + forward_vector[2] * 10]

        view_matrix = pybullet.computeViewMatrix(
            pos_world,
            camera_target,
            up_vector,
            physicsClientId=self.physics_client)

        with self.resolution_lock:
            camera_image = pybullet.getCameraImage(
                self.resolution.width,
                self.resolution.height,
                view_matrix,
                self.projection_matrix,
                renderer=pybullet.ER_BULLET_HARDWARE_OPENGL,
                flags=pybullet.ER_NO_SEGMENTATION_MASK,
                physicsClientId=self.physics_client)

        return camera_image

    def _resetActiveCamera(self):
        """
        INTERNAL METHOD, called when unsubscribing from the active camera, when
        Python is exitted or when the SimulationManager resets/stops a
        simulation instance
        """
        Camera.ACTIVE_CAMERA_ID[self.physics_client] = -1

        if self.extraction_thread.isAlive():
            self.extraction_thread.join()

    def _waitForCorrectImageFormat(self):
        """
        INTERNAL METHOD, to be called after the launch of the exctraction
        thread. Blocking method, that will return when the array retrieved from
        the @getCameraImage is not None and corresponds to the current
        resolution
        """
        try:
            assert self.extraction_thread.isAlive()

            while self.getFrame() is None:
                continue

            while True:
                image = self.getFrame()
                if image.shape[1] == self.resolution.width and\
                   image.shape[0] == self.resolution.height:
                    break

        except AssertionError:
            return


class CameraRgb(Camera):
    """
    Class representing a virtual rgb camera
    """

    def __init__(
            self,
            robot_model,
            link,
            physicsClientId=0,
            resolution=Camera.K_QVGA):
        """
        Constructor

        Parameters:
            robot_model - the pybullet model of the robot
            link - The link (Link type) onto which the camera's attached
            physicsClientId - The id of the simulated instance in which the
            camera is to be spawned
            resolution - The resolution of the camera
        """
        Camera.__init__(
            self,
            robot_model,
            link,
            physicsClientId=physicsClientId)

        self._setFov(67.4)
        self.rgb_image = None

    def subscribe(self, resolution=Camera.K_QVGA):
        """
        Overload @subscribe from @Camera

        Parameters:
            resolution - CameraResolution object, the resolution of the camera.
            By default, the resolution is QVGA
        """
        Camera.subscribe(self, id(self), resolution)
        self.extraction_thread =\
            threading.Thread(target=self._frameExtractionLoop)
        self.extraction_thread.start()
        self._waitForCorrectImageFormat()

    def unsubscribe(self):
        """
        Overload @unsubscribe from @Camera
        """
        Camera.unsubscribe(self, id(self))

    def _frameExtractionLoop(self):
        """
        Frame extraction loop, has to be threaded. The resolution and the FOV
        have to be specified beforehand
        """
        rgb_image = np.zeros((
            self.resolution.height,
            self.resolution.width,
            3))

        try:
            while True:
                assert id(self) == Camera.ACTIVE_CAMERA_ID[self.physics_client]
                camera_image = self._getCameraImage()

                camera_image = np.reshape(
                    camera_image[2],
                    (camera_image[1], camera_image[0], 4))

                rgb_image[:, :, 0] =\
                    (1 - camera_image[:, :, 3]) * camera_image[:, :, 2] +\
                    camera_image[:, :, 3] * camera_image[:, :, 2]

                rgb_image[:, :, 1] =\
                    (1 - camera_image[:, :, 3]) * camera_image[:, :, 1] +\
                    camera_image[:, :, 3] * camera_image[:, :, 1]

                rgb_image[:, :, 2] =\
                    (1 - camera_image[:, :, 3]) * camera_image[:, :, 0] +\
                    camera_image[:, :, 3] * camera_image[:, :, 0]

                rgb_image = rgb_image.astype(np.uint8)

                with self.frame_lock:
                    self.frame = rgb_image.copy()

        except AssertionError:
            return


class CameraDepth(Camera):
    """
    Class representing a virtual depth camera
    """

    def __init__(
            self,
            robot_model,
            link,
            physicsClientId=0,
            resolution=Camera.K_QVGA):
        """
        Constructor

        Parameters:
            robot_model - the pybullet model of the robot
            link - The link (Link type) onto which the camera's attached
            physicsClientId - The id of the simulated instance in which the
            camera is to be spawned
            resolution - The resolution of the camera
        """
        Camera.__init__(
            self,
            robot_model,
            link,
            physicsClientId=physicsClientId)

        self._setFov(70)

    def subscribe(self, resolution=Camera.K_QVGA):
        """
        Overload @subscribe from @Camera

        Parameters:
            resolution - CameraResolution object, the resolution of the camera.
            By default, the resolution is QVGA
        """
        Camera.subscribe(self, id(self), resolution)
        self.extraction_thread =\
            threading.Thread(target=self._frameExtractionLoop)
        self.extraction_thread.start()
        self._waitForCorrectImageFormat()

    def unsubscribe(self):
        """
        Overload @unsubscribe from @Camera
        """
        Camera.unsubscribe(self, id(self))

    def _frameExtractionLoop(self):
        """
        Frame extraction loop, has to be threaded. The resolution and the FOV
        have to be specified beforehand
        """
        far = self.far_plane / 10
        near = self.near_plane / 10
        try:
            while True:
                assert id(self) == Camera.ACTIVE_CAMERA_ID[self.physics_client]
                camera_image = self._getCameraImage()
                depth_image = camera_image[3]

                depth_image = (far * near) /\
                    (far - (far - near) * depth_image)

                image_histogram, bins = np.histogram(
                    depth_image.flatten(),
                    256,
                    normed=True)

                # Cumulative distribution function
                cdf = image_histogram.cumsum()
                # Normalize for 16 bits
                cdf = 65535 * cdf / cdf[-1]

                # Use linear interpolation of cdf to find new pixel values
                image_equalized = np.interp(
                    depth_image.flatten(),
                    bins[:-1],
                    cdf)
                depth_image = image_equalized.reshape(depth_image.shape)

                with self.frame_lock:
                    self.frame = depth_image.astype(np.uint16)

        except AssertionError:
            return



#!/usr/bin/env python
# coding: utf-8

import sys
IS_VERSION_PYTHON_3 = sys.version_info[0] >= 3


class Link:
    """
    Class describing a robot link
    """

    def __init__(self, joint_info):
        """
        Constructor

        Parameters:
            name - The name of the joint
            joint_info - Informations returned by the getJointInfo API
        """
        self.index = joint_info[0]
        self.name = joint_info[12]
        self.parent_index = joint_info[16]
        if IS_VERSION_PYTHON_3:
            self.name.decode('utf-8')

    def getIndex(self):
        """
        Getter for the index parameter

        Returns:
            index - The link's index
        """
        return self.index

    def getName(self):
        """
        Getter for the name parameter

        Returns:
            name - The link's name
        """
        return self.name

    def getParentIndex(self):
        """
        Getter for the parent index parameter

        Returns:
            parent_index - The index of the link's parent
        """
        return self.parent_index



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="qibullet",
    version="0.0.6",
    author="Maxime Busy, Maxime Caniot",
    author_email="",
    description="Bullet-based simulation for SoftBank Robotics' robots",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ProtolabSBRE/qibullet",
    packages=setuptools.find_packages(),
    install_requires=['numpy', 'pybullet'],
    package_data={"qibullet": [
        "robot_data/pepper_1.7/*.urdf",
        "robot_data/pepper_1.7/meshes/*.obj",
        "robot_data/pepper_1.7/meshes/*.mtl",
        "robot_data/pepper_1.7/meshes/*.stl",
        "robot_data/pepper_1.7/meshes/*.png",
        "robot_data/nao_V40/*.urdf",
        "robot_data/nao_V40/meshes/*.obj",
        "robot_data/nao_V40/meshes/*.mtl",
        "robot_data/nao_V40/meshes/*.stl",
        "robot_data/nao_V40/meshes/*.png"]},
    keywords=[
        'physics simulation',
        'robotics',
        'naoqi',
        'softbank',
        'pepper',
        'nao',
        'romeo',
        'robot'],
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        "Operating System :: Microsoft",
        'Topic :: Games/Entertainment :: Simulation',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Framework :: Robot Framework :: Tool'
    ]
)



from qibullet.link import Link
from qibullet.joint import Joint
from qibullet.pepper_virtual import PepperVirtual
from qibullet.simulation_manager import SimulationManager
from qibullet.camera import Camera, CameraRgb, CameraDepth, CameraResolution

try:
    from qibullet.ros_wrapper import PepperRosWrapper

except ImportError:
    pass

name = 'qibullet'



#!/usr/bin/env python
# coding: utf-8
import math
import pybullet


def getDistance(point_a, point_b):
    [x1, y1, z1] = point_a
    [x2, y2, z2] = point_b
    return math.sqrt((x2 - x1)**2 + (y2 - y1)**2 + (z2 - z1)**2)


def getOrientation(quaternion_a, quaternion_b):
    theta_a = pybullet.getEulerFromQuaternion(quaternion_a)
    theta_b = pybullet.getEulerFromQuaternion(quaternion_b)
    return theta_b[-1] - theta_a[-1]


def computeVelocity(acc, vel_min, vel_max, dist_traveled, dist_remained):
    distance_acc = (vel_max * vel_max) / (2 * acc)
    if dist_traveled < distance_acc:
        vel_computed = (vel_max - vel_min) *\
            dist_traveled / distance_acc + vel_min
    if dist_traveled >= distance_acc:
        vel_computed = vel_max
    if dist_remained < distance_acc:
        vel_computed = (vel_max - vel_min) *\
            dist_remained / distance_acc + vel_min
    return vel_computed
